from django.db.models.functions import (
    SHA224, SHA256, SHA384, SHA512, Ceil, Collate, ConcatPair, Random,
    StrIndex,
)


def ceil(self, compiler, connection, **extra_context):
    return self.as_sql(compiler, connection, function='CEIL', **extra_context)


def collate(self, compiler, connection, **extra_context):
    # https://docs.snowflake.com/en/sql-reference/functions/collate.html
    return self.as_sql(
        compiler,
        connection,
        # Snowflake requires single quotes instead of double quotes.
        collation=f"'{self.collation}'",
        # COLLATE(<string_expression>, '<collation_specification>')
        template='%(function)s(%(expressions)s, %(collation)s)',
        **extra_context,
    )


def concatpair(self, compiler, connection, **extra_context):
    # coalesce() prevents Concat from returning null instead of empty string.
    return self.coalesce().as_sql(compiler, connection, **extra_context)


def random(self, compiler, connection, **extra_context):
    template = 'UNIFORM(0, 0.99999999999999999, RANDOM())'
    return self.as_sql(compiler, connection, template=template, **extra_context)


def strindex(self, compiler, connection, **extra_context):
    # POSITION takes arguments in the opposite order of other databases.
    # https://docs.snowflake.com/en/sql-reference/functions/position.html
    return StrIndex(
        self.source_expressions[1],
        self.source_expressions[0],
    ).as_sql(compiler, connection, function='POSITION', **extra_context)


def register_functions():
    SHA224.as_snowflake = SHA224.as_mysql
    SHA256.as_snowflake = SHA256.as_mysql
    SHA384.as_snowflake = SHA384.as_mysql
    SHA512.as_snowflake = SHA512.as_mysql
    Ceil.as_snowflake = ceil
    Collate.as_snowflake = collate
    ConcatPair.as_snowflake = concatpair
    Random.as_snowflake = random
    StrIndex.as_snowflake = strindex



from itertools import chain

from django.db.models import JSONField
from django.db.models.sql import compiler


class SQLInsertCompiler(compiler.SQLInsertCompiler):
    def as_sql(self):
        # We don't need quote_name_unless_alias() here, since these are all
        # going to be column names (so we can avoid the extra overhead).
        qn = self.connection.ops.quote_name
        opts = self.query.get_meta()
        insert_statement = self.connection.ops.insert_statement(
            on_conflict=self.query.on_conflict,
        )
        result = ["%s %s" % (insert_statement, qn(opts.db_table))]
        fields = self.query.fields or [opts.pk]
        result.append("(%s)" % ", ".join(qn(f.column) for f in fields))

        select_columns = []
        if self.query.fields:
            value_rows = [
                [
                    self.prepare_value(field, self.pre_save_val(field, obj))
                    for field in fields
                ]
                for obj in self.query.objs
            ]
            has_json_field = False
            for i, field in enumerate(fields, 1):
                if isinstance(field, JSONField):
                    has_json_field = True
                    select_columns.append(f'parse_json(${i})')
                else:
                    select_columns.append(f'${i}')
            if not has_json_field:
                select_columns = []
        else:
            # An empty object.
            value_rows = [
                [self.connection.ops.pk_default_value()] for _ in self.query.objs
            ]
            fields = [None]

        # Currently the backends just accept values when generating bulk
        # queries and generate their own placeholders. Doing that isn't
        # necessary and it should be possible to use placeholders and
        # expressions in bulk inserts too.
        can_bulk = (
            not self.returning_fields and self.connection.features.has_bulk_insert
        )

        placeholder_rows, param_rows = self.assemble_as_sql(fields, value_rows)

        on_conflict_suffix_sql = self.connection.ops.on_conflict_suffix_sql(
            fields,
            self.query.on_conflict,
            (f.column for f in self.query.update_fields),
            (f.column for f in self.query.unique_fields),
        )
        if (
            self.returning_fields
            and self.connection.features.can_return_columns_from_insert
        ):
            if self.connection.features.can_return_rows_from_bulk_insert:
                result.append(
                    self.connection.ops.bulk_insert_sql(fields, placeholder_rows)
                )
                params = param_rows
            else:
                result.append("VALUES (%s)" % ", ".join(placeholder_rows[0]))
                params = [param_rows[0]]
            if on_conflict_suffix_sql:
                result.append(on_conflict_suffix_sql)
            # Skip empty r_sql to allow subclasses to customize behavior for
            # 3rd party backends. Refs #19096.
            r_sql, self.returning_params = self.connection.ops.return_insert_columns(
                self.returning_fields
            )
            if r_sql:
                result.append(r_sql)
                params += [self.returning_params]
            return [(" ".join(result), tuple(chain.from_iterable(params)))]

        if select_columns:
            result.append('SELECT ' + (", ".join(c for c in select_columns)) + ' FROM')

        if can_bulk:
            result.append(self.connection.ops.bulk_insert_sql(fields, placeholder_rows))
            if on_conflict_suffix_sql:
                result.append(on_conflict_suffix_sql)
            return [(" ".join(result), tuple(p for ps in param_rows for p in ps))]
        else:
            if on_conflict_suffix_sql:
                result.append(on_conflict_suffix_sql)
            return [
                (" ".join(result + ["VALUES (%s)" % ", ".join(p)]), vals)
                for p, vals in zip(placeholder_rows, param_rows)
            ]


SQLCompiler = compiler.SQLCompiler
SQLDeleteCompiler = compiler.SQLDeleteCompiler
SQLUpdateCompiler = compiler.SQLUpdateCompiler
SQLAggregateCompiler = compiler.SQLAggregateCompiler



from django.db import NotSupportedError
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.models import NOT_PROVIDED


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    sql_create_column_inline_fk = (
        'CONSTRAINT %(name)s FOREIGN KEY REFERENCES %(to_table)s(%(to_column)s)'
    )
    sql_delete_procedure = 'DROP PROCEDURE %(procedure)s(%(param_types)s)'

    def _create_index_sql(self, model, fields=None, **kwargs):
        # Snowflake doesn't use indexes.
        return ''

    def _model_indexes_sql(self, model):
        return []

    def _field_indexes_sql(self, model, field):
        return []

    def add_index(self, model, index):
        pass

    def remove_index(self, model, index):
        pass

    def alter_index_together(self, model, old_index_together, new_index_together):
        pass

    def add_field(self, model, field):
        # Special-case implicit M2M tables
        if field.many_to_many and field.remote_field.through._meta.auto_created:
            return self.create_model(field.remote_field.through)
        # Get the column's definition
        definition, params = self.column_sql(model, field, exclude_not_null=True)
        # It might not actually have a column behind it
        if definition is None:
            return
        if col_type_suffix := field.db_type_suffix(connection=self.connection):
            definition += f" {col_type_suffix}"
        if field.remote_field and field.db_constraint:
            # Add FK constraint inline.
            constraint_suffix = '_fk_%(to_table)s_%(to_column)s'
            to_table = field.remote_field.model._meta.db_table
            to_column = field.remote_field.model._meta.get_field(field.remote_field.field_name).column
            definition += " " + self.sql_create_column_inline_fk % {
                'name': self._fk_constraint_name(model, field, constraint_suffix),
                'column': self.quote_name(field.column),
                'to_table': self.quote_name(to_table),
                'to_column': self.quote_name(to_column),
            }
        # Build the SQL and run it
        sql = self.sql_create_column % {
            "table": self.quote_name(model._meta.db_table),
            "column": self.quote_name(field.column),
            "definition": definition,
        }
        self.execute(sql, params)
        # Set default values on existing rows. Django usually uses database
        # defaults for this, but Snowflake doesn't allow dropping a default
        # value for columns added after the table is created.
        effective_default = self.effective_default(field)
        if effective_default is not None and field.db_default is NOT_PROVIDED:
            self.execute(
                "UPDATE %(table)s SET %(column)s=%%s" % {
                    "table": self.quote_name(model._meta.db_table),
                    "column": self.quote_name(field.column),
                },
                (effective_default,),
            )
        # Add NOT NULL to the column, if required, after it's created rather
        # when it's created, otherwise a database default would be required
        # which can't be dropped as discussed in the previous comment.
        if not field.null:
            db_params = field.db_parameters(connection=self.connection)
            self.execute(
                self.sql_alter_column % {
                    "table": self.quote_name(model._meta.db_table),
                    "changes": self.sql_alter_column_not_null % {
                        "column": self.quote_name(field.column),
                        "type": db_params["type"],
                    },
                },
            )

    def column_sql(self, model, field, include_default=False, exclude_not_null=False):
        params = []
        # Get the column's type and use that as the basis of the SQL
        db_params = field.db_parameters(connection=self.connection)
        sql = db_params["type"]
        # Check for fields that aren't actually columns (e.g. M2M)
        if sql is None:
            return None, None
        collation = db_params.get('collation')
        if collation:
            sql += self._collate_sql(collation)
        if field.generated:
            sql += self._column_generated_sql(field)
        if not field.null and not exclude_not_null:
            sql += " NOT NULL"
        # Add database default.
        if field.db_default is not NOT_PROVIDED:
            default_sql, default_params = self.db_default_sql(field)
            sql += f" DEFAULT {default_sql}"
            params.extend(default_params)
        if field.primary_key:
            sql += " PRIMARY KEY"
        if field.unique:
            sql += " UNIQUE"
        if field.db_comment:
            sql += self._comment_sql(field.db_comment)
        return sql, params

    def _collate_sql(self, collation, old_collation=None, table_name=None):
        # Collation must be single quoted instead of double quoted.
        return f" COLLATE '{collation}'" if collation else ""

    def _comment_sql(self, comment):
        comment_sql = super()._comment_sql(comment)
        return f" COMMENT {comment_sql}"

    def _alter_column_comment_sql(self, model, new_field, new_type, new_db_comment):
        return (
            self.sql_alter_column_comment
            % {
                "table": self.quote_name(model._meta.db_table),
                "column": self.quote_name(new_field.column),
                # Avoid "COMMENT " prefix.
                "comment": BaseDatabaseSchemaEditor._comment_sql(self, new_db_comment),
            },
            [],
        )

    def _alter_field(self, model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params, strict=False):
        super()._alter_field(
            model, old_field, new_field, old_type, new_type,
            old_db_params, new_db_params, strict,
        )
        auto_fields = self.connection.data_types_suffix
        old_internal_type = old_field.get_internal_type()
        new_internal_type = new_field.get_internal_type()
        # Altering to an AutoField isn't supported because Snowflake doesn't
        # support "ALTER COLUMN... SET DEFAULT" which would be need to add
        # a sequence to the column.
        if old_internal_type not in auto_fields and new_internal_type in auto_fields:
            raise NotSupportedError(
                "Changing field %(field_name)s to %(field_type)s isn't supported." % {
                    'field_name': old_field.name,
                    'field_type': new_internal_type,
                }
            )
        # If migrating away from AutoField, drop AUTOINCREMENT.
        if old_internal_type in auto_fields and new_internal_type not in auto_fields:
            self.execute(self.sql_alter_column % {
                "table": self.quote_name(model._meta.db_table),
                "changes": self.sql_alter_column_no_default % {
                    "column": self.quote_name(new_field.column),
                },
            })

    def _column_generated_sql(self, field):
        """Return the SQL to use in a GENERATED ALWAYS clause."""
        expression_sql, params = field.generated_sql(self.connection)
        if params:
            expression_sql = expression_sql % tuple(self.quote_value(p) for p in params)
        return f" AS ({expression_sql})"

    def quote_value(self, value):
        if isinstance(value, str):
            return "'%s'" % value.replace("'", "\\'")
        else:
            return str(value)

    def prepare_default(self, value):
        return self.quote_value(value).replace("%", "%%")

    def skip_default_on_alter(self, field):
        # Snowflake: Unsupported feature 'Alter Column Set Default'.
        return True



import signal

from django.db.backends.base.client import BaseDatabaseClient


class DatabaseClient(BaseDatabaseClient):
    executable_name = 'snowsql'

    @classmethod
    def settings_to_cmd_args_env(cls, settings_dict, parameters):
        args = [cls.executable_name]

        account = settings_dict.get('ACCOUNT')
        dbname = settings_dict.get('NAME')
        host = settings_dict.get('HOST')
        password = settings_dict.get('PASSWORD')
        schema = settings_dict.get('SCHEMA')
        user = settings_dict.get('USER')
        warehouse = settings_dict.get('WAREHOUSE')
        # snowflake.connector.connect() parameters that have a corresponding
        # snowsql option.
        options = settings_dict['OPTIONS']
        authenticator = options.get('authenticator')
        client_session_keep_alive = options.get('client_session_keep_alive')
        passcode = options.get('passcode')
        passcode_in_password = options.get('passcode_in_password')
        private_key_file = options.get('private_key_file')
        private_key_file_pwd = options.get('private_key_file_pwd')
        role = options.get('role')
        token = options.get('token')

        if account:
            args += ['-a', account]
        if authenticator:
            args += ['--authenticator', authenticator]
        if client_session_keep_alive:
            args += ['--client-session-keep-alive']
        if dbname:
            args += ['-d', dbname]
        if host:
            args += ['-h', host]
        if passcode:
            args += ['--mfa-passcode', passcode]
        if passcode_in_password:
            args += ['--mfa-passcode-in-password']
        if private_key_file:
            args += ['--private-key-path', private_key_file]
        if role:
            args += ['-r', role]
        if schema:
            args += ['-s', schema]
        if token:
            args += ['--token', token]
        if user:
            args += ['-u', user]
        if warehouse:
            args += ['-w', warehouse]

        env = {}
        if password:
            env['SNOWSQL_PWD'] = password
        if private_key_file_pwd:
            env['SNOWSQL_PRIVATE_KEY_PASSPHRASE'] = private_key_file_pwd
        return args, (env or None)

    def runshell(self, parameters):
        sigint_handler = signal.getsignal(signal.SIGINT)
        try:
            # Allow SIGINT to pass to snowsql to abort queries.
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            super().runshell(parameters)
        finally:
            # Restore the original SIGINT handler.
            signal.signal(signal.SIGINT, sigint_handler)



import os

from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.base import BaseDatabaseWrapper
from django.utils.asyncio import async_unsafe

try:
    import snowflake.connector as Database
except ImportError as e:
    raise ImproperlyConfigured("Error loading snowflake connector module: %s" % e)

# Some of these import snowflake connector, so import them after checking if it's installed.
from . import __version__                                   # NOQA isort:skip
from .client import DatabaseClient                          # NOQA isort:skip
from .creation import DatabaseCreation                      # NOQA isort:skip
from .features import DatabaseFeatures                      # NOQA isort:skip
from .introspection import DatabaseIntrospection            # NOQA isort:skip
from .operations import DatabaseOperations                  # NOQA isort:skip
from .schema import DatabaseSchemaEditor                    # NOQA isort:skip


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = 'snowflake'
    display_name = 'Snowflake'
    data_types = {
        'AutoField': 'NUMBER(10,0)',
        'BigAutoField': 'NUMBER(38,0)',
        'BinaryField': 'BINARY',
        'BooleanField': 'BOOLEAN',
        'CharField': 'VARCHAR(%(max_length)s)',
        'DateField': 'DATE',
        'DateTimeField': 'TIMESTAMP_LTZ',
        'DecimalField': 'NUMBER(%(max_digits)s,%(decimal_places)s)',
        'DurationField': 'NUMBER(38,0)',
        'FileField': 'VARCHAR(%(max_length)s)',
        'FilePathField': 'VARCHAR(%(max_length)s)',
        'FloatField': 'FLOAT',
        'IntegerField': 'NUMBER(10,0)',
        'BigIntegerField': 'NUMBER(38,0)',
        'GenericIPAddressField': 'VARCHAR(39)',
        'JSONField': 'VARIANT',
        'NullBooleanField': 'BOOLEAN',
        'OneToOneField': 'NUMBER(38,0)',
        'PositiveBigIntegerField': 'NUMBER(38,0)',
        'PositiveIntegerField': 'NUMBER(10,0)',
        'PositiveSmallIntegerField': 'NUMBER(5,0)',
        'SlugField': 'VARCHAR(%(max_length)s)',
        'SmallAutoField': 'NUMBER(5,0)',
        'SmallIntegerField': 'NUMBER(5,0)',
        'TextField': 'VARCHAR',
        'TimeField': 'TIME',
        'UUIDField': 'VARCHAR(32)',
    }
    data_types_suffix = {
        'AutoField': 'AUTOINCREMENT',
        'BigAutoField': 'AUTOINCREMENT',
        'SmallAutoField': 'AUTOINCREMENT',
    }
    operators = {
        'exact': '= %s',
        'iexact': "ILIKE %s ESCAPE '\\\\'",
        'contains': "LIKE %s ESCAPE '\\\\'",
        'icontains': "ILIKE %s ESCAPE '\\\\'",
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': "LIKE %s ESCAPE '\\\\'",
        'endswith': "LIKE %s ESCAPE '\\\\'",
        'istartswith': "ILIKE %s ESCAPE '\\\\'",
        'iendswith': "ILIKE %s ESCAPE '\\\\'",
    }
    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\\', '\\\\'), '%%', '\\%%'), '_', '\\_')"
    pattern_ops = {
        'contains': "LIKE '%%' || {} || '%%' ESCAPE '\\\\'",
        'icontains': "ILIKE '%%' || {} || '%%' ESCAPE '\\\\'",
        'startswith': "LIKE {} || '%%' ESCAPE '\\\\'",
        'istartswith': "ILIKE {} || '%%' ESCAPE '\\\\'",
        'endswith': "LIKE '%%' || {} ESCAPE '\\\\'",
        'iendswith': "ILIKE '%%' || {} ESCAPE '\\\\'",
    }

    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor

    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations

    password_not_required_options = ('private_key', 'private_key_file', 'authenticator')
    settings_is_missing = "settings.DATABASES is missing '%s' for 'django_snowflake'."

    def get_connection_params(self):
        settings_dict = self.settings_dict
        conn_params = {
            'interpolate_empty_sequences':  True,
            **settings_dict['OPTIONS'],
        }
        if os.environ.get('RUNNING_DJANGOS_TEST_SUITE') != 'true':
            conn_params['application'] = 'Django_SnowflakeConnector_%s' % __version__

        if settings_dict['NAME']:
            conn_params['database'] = self.ops.quote_name(settings_dict['NAME'])

        if settings_dict['USER']:
            conn_params['user'] = settings_dict['USER']
        else:
            raise ImproperlyConfigured(self.settings_is_missing % 'USER')

        if settings_dict['PASSWORD']:
            conn_params['password'] = settings_dict['PASSWORD']
        elif all(x not in conn_params for x in self.password_not_required_options):
            raise ImproperlyConfigured(self.settings_is_missing % 'PASSWORD')

        if settings_dict.get('ACCOUNT'):
            conn_params['account'] = settings_dict['ACCOUNT']
        else:
            raise ImproperlyConfigured(self.settings_is_missing % 'ACCOUNT')

        if settings_dict.get('WAREHOUSE'):
            conn_params['warehouse'] = self.ops.quote_name(settings_dict['WAREHOUSE'])
        else:
            raise ImproperlyConfigured(self.settings_is_missing % 'WAREHOUSE')

        if settings_dict.get('SCHEMA'):
            conn_params['schema'] = self.ops.quote_name(settings_dict['SCHEMA'])
        else:
            raise ImproperlyConfigured(self.settings_is_missing % 'SCHEMA')

        return conn_params

    @async_unsafe
    def get_new_connection(self, conn_params):
        return Database.connect(**conn_params)

    def ensure_timezone(self):
        if self.connection is None:
            return False
        with self.connection.cursor() as cursor:
            conn_timezone_name = cursor.execute("SHOW PARAMETERS LIKE 'TIMEZONE'").fetchone()[1]
        timezone_name = self.timezone_name
        if timezone_name and conn_timezone_name != timezone_name:
            with self.connection.cursor() as cursor:
                cursor.execute("ALTER SESSION SET TIMEZONE=%s", [timezone_name])
            return True
        return False

    def init_connection_state(self):
        # AUTOINCREMENT IDs must be monotonically increasing in order for
        # DatabaseOperations.last_insert_id() to fetch the correct ID.
        with self.connection.cursor() as cursor:
            cursor.execute("ALTER SESSION SET NOORDER_SEQUENCE_AS_DEFAULT=False")
        timezone_changed = self.ensure_timezone()
        if timezone_changed:
            # Commit after setting the time zone (see #17062)
            # (This is copied from the postgresql backend.)
            if not self.get_autocommit():
                self.connection.commit()

    @async_unsafe
    def create_cursor(self, name=None):
        return self.connection.cursor()

    def _set_autocommit(self, autocommit):
        with self.wrap_database_errors:
            self.connection.autocommit(autocommit)

    def is_usable(self):
        try:
            # Use a cursor directly, bypassing Django's utilities.
            with self.connection.cursor() as cursor:
                cursor.execute('SELECT current_version()')
        except Database.Error:
            return False
        else:
            return True



from collections import namedtuple

from django.db.backends.base.introspection import (
    BaseDatabaseIntrospection, FieldInfo as BaseFieldInfo,
    TableInfo as BaseTableInfo,
)
from django.utils.regex_helper import _lazy_re_compile

FieldInfo = namedtuple('FieldInfo', BaseFieldInfo._fields + ('pk', 'comment'))
TableInfo = namedtuple('TableInfo', BaseTableInfo._fields + ('comment',))
collation_re = _lazy_re_compile(r"^VARCHAR\(\d+\) COLLATE '([\w+\-]+)'$")
field_size_re = _lazy_re_compile(r'^[A-Z]+\((\d+)\)')
precision_and_scale_re = _lazy_re_compile(r'^NUMBER\((\d+),(\d+)\)$')


def get_collation(name):
    """
    Return the collation from a "VARCHAR(11) COLLATE 'collation'" type name.
    """
    m = collation_re.search(name)
    return m[1] if m else None


def get_data_type(name):
    """Extract the data type name from an ABC(...) type name."""
    return name.split('(', 1)[0]


def get_field_size(name):
    """Extract the size number from a VARCHAR(11) type name."""
    m = field_size_re.search(name)
    return int(m[1]) if m else None


def get_precision_and_scale(name):
    """
    Return the precision (first number) and scale (second number) from a
    NUMBER(38,0) type name.
    """
    m = precision_and_scale_re.search(name)
    return (int(m[1]), int(m[2])) if m else (None, None)


class DatabaseIntrospection(BaseDatabaseIntrospection):
    # Maps Snowflake data types returned by `DESCRIBE TABLE` to Django Fields.
    data_types_reverse = {
        'BINARY': 'BinaryField',
        'BOOLEAN': 'BooleanField',
        'DATE': 'DateField',
        'FLOAT': 'FloatField',
        'NUMBER': 'BigIntegerField',
        'TIME': 'TimeField',
        'TIMESTAMP_LTZ': 'DateTimeField',
        'VARCHAR': 'CharField',
        'VARIANT': 'JSONField',
    }

    def get_constraints(self, cursor, table_name):
        table_name = self.connection.ops.quote_name(table_name)
        constraints = {}
        # Foreign keys
        cursor.execute(f'SHOW IMPORTED KEYS IN TABLE {table_name}')
        for row in cursor.fetchall():
            constraints[self.identifier_converter(row[12])] = {
                'columns': [self.identifier_converter(row[8])],
                'primary_key': False,
                'unique': False,
                'foreign_key': (self.identifier_converter(row[3]), self.identifier_converter(row[4])),
                'check': False,
                'index': False,
            }
        # Primary keys
        cursor.execute(f'SHOW PRIMARY KEYS IN TABLE {table_name}')
        for row in cursor.fetchall():
            constraints[self.identifier_converter(row[6])] = {
                'columns': [self.identifier_converter(row[4])],
                'primary_key': True,
                'unique': False,
                'foreign_key': None,
                'check': False,
                'index': False,
            }
        # Unique constraints
        cursor.execute(f'SHOW UNIQUE KEYS IN TABLE {table_name}')
        # The columns of multi-column unique indexes are ordered by row[5].
        # Map {constraint_name: [(row[5], column_name), ...] so the columns can
        # be sorted for each constraint.
        unique_column_orders = {}
        for row in cursor.fetchall():
            column_name = self.identifier_converter(row[4])
            constraint_name = self.identifier_converter(row[6])
            if constraint_name in constraints:
                # If the constraint name is already present, this is a
                # multi-column unique constraint.
                constraints[constraint_name]['columns'].append(column_name)
                unique_column_orders[constraint_name].append((row[5], column_name))
            else:
                constraints[constraint_name] = {
                    'columns': [column_name],
                    'primary_key': False,
                    'unique': True,
                    'foreign_key': None,
                    'check': False,
                    'index': False,
                }
                unique_column_orders[constraint_name] = [(row[5], column_name)]
        # Order the columns of multi-column unique indexes.
        for constraint_name, orders in unique_column_orders.items():
            constraints[constraint_name]['columns'] = [col for _, col in sorted(orders)]
        return constraints

    def get_primary_key_column(self, cursor, table_name):
        pks = [field.name for field in self.get_table_description(cursor, table_name) if field.pk]
        return pks[0] if pks else None

    def get_relations(self, cursor, table_name):
        """
        Return a dictionary of {field_name: (field_name_other_table, other_table)}
        representing all foreign keys in the given table.
        """
        table_name = self.connection.ops.quote_name(table_name)
        cursor.execute(f'SHOW IMPORTED KEYS IN TABLE {table_name}')
        return {
            self.identifier_converter(row[8]): (self.identifier_converter(row[4]), self.identifier_converter(row[3]))
            for row in cursor.fetchall()
        }

    def get_field_type(self, data_type, description):
        field_type = super().get_field_type(data_type, description)
        # 16777216 is the default size if max_length isn't specified.
        if data_type == 'VARCHAR' and description.display_size == 16777216:
            return 'TextField'
        # Handle NUMBER if it's something besides BigAutoField.
        if data_type == 'NUMBER':
            if description.scale != 0:
                return 'DecimalField'
            elif description.precision == 5:
                field_type = 'SmallIntegerField'
            elif description.precision == 10:
                field_type = 'IntegerField'
        # Handle AutoField and variants.
        if description.default and 'IDENTITY' in description.default:
            if field_type == 'IntegerField':
                return 'AutoField'
            elif field_type == 'BigIntegerField':
                return 'BigAutoField'
            elif field_type == 'SmallIntegerField':
                return 'SmallAutoField'
        return field_type

    def get_table_description(self, cursor, table_name):
        cursor.execute('DESCRIBE TABLE %s' % self.connection.ops.quote_name(table_name))
        table_info = cursor.fetchall()
        return [
            FieldInfo(
                self.identifier_converter(name),  # name
                get_data_type(data_type),  # type_code
                get_field_size(data_type),  # display_size
                None,  # internal_size
                *get_precision_and_scale(data_type),  # precision, scale
                null == 'Y',  # null_ok
                default,  # default
                get_collation(data_type),  # collation
                pk == 'Y',  # pk
                comment,  # comment
            )
            for (
                name, data_type, kind, null, default, pk, unique_key, check,
                # *_ ignores policy_name, privacy_domain, and any future
                # columns in DESCRIBE TABLE output.
                expression, comment, *_
            ) in table_info
        ]

    def identifier_converter(self, name):
        # Add quotes around Snowflake generated constraint names like
        # SYS_CONSTRAINT_e8775210-b2d4-4947-b382-c57cecc6bb6d to preserve
        # casing.
        if name.startswith("SYS_CONSTRAINT_"):
            return f'"{name}"'
        # TODO: If the identifier field isn't uppercase, then it needs to be
        # quoted to preserve its case.
        # https://github.com/Snowflake-Labs/django-snowflake/issues/43
        # This may require some changes in Django itself to work properly. This
        # would replace the special handling of SYS_CONSTRAINT_ above.
        # if name != name.upper():
        #    return f'"{name}"'
        # Otherwise, the identifier name can be lowercased for use as a model
        # field name, for example. DatabaseOperations.quote_name() reverses
        # this transformation by uppercasing the name.
        return name.lower()

    def get_table_list(self, cursor):
        cursor.execute('SHOW TABLES')
        tables = [
            TableInfo(
                self.identifier_converter(row[1]),  # table name
                't',  # 't' for table
                row[5],  # comment
            ) for row in cursor.fetchall()
        ]
        cursor.execute('SHOW VIEWS')
        views = [
            TableInfo(
                self.identifier_converter(row[1]),   # view name
                'v',  # 'v' for view
                row[6],  # comment
            ) for row in cursor.fetchall()
        ]
        return tables + views



import decimal
import uuid

from django.conf import settings
from django.db.backends.base.operations import BaseDatabaseOperations
from django.utils import timezone


class DatabaseOperations(BaseDatabaseOperations):
    cast_char_field_without_max_length = 'varchar'
    cast_data_types = {
        'AutoField': 'NUMBER',
        'BigAutoField': 'NUMBER',
        'SmallAutoField': 'NUMBER',
    }
    compiler_module = 'django_snowflake.compiler'
    explain_prefix = 'EXPLAIN USING'

    def bulk_insert_sql(self, fields, placeholder_rows):
        placeholder_rows_sql = (', '.join(row) for row in placeholder_rows)
        values_sql = ', '.join('(%s)' % sql for sql in placeholder_rows_sql)
        return 'VALUES ' + values_sql

    def combine_expression(self, connector, sub_expressions):
        lhs, rhs = sub_expressions
        if connector == '&':
            return 'BITAND(%s)' % ','.join(sub_expressions)
        elif connector == '|':
            return 'BITOR(%(lhs)s,%(rhs)s)' % {'lhs': lhs, 'rhs': rhs}
        elif connector == '#':
            return 'BITXOR(%(lhs)s, %(rhs)s)' % {'lhs': lhs, 'rhs': rhs}
        elif connector == '<<':
            return 'BITSHIFTLEFT(%(lhs)s, %(rhs)s)' % {'lhs': lhs, 'rhs': rhs}
        elif connector == '>>':
            return 'BITSHIFTRIGHT(%(lhs)s, %(rhs)s)' % {'lhs': lhs, 'rhs': rhs}
        elif connector == '^':
            return 'POWER(%s)' % ','.join(sub_expressions)
        return super().combine_expression(connector, sub_expressions)

    def _convert_sql_to_tz(self, sql, params, tzname):
        if tzname and settings.USE_TZ:
            return f"CONVERT_TIMEZONE(%s, TO_TIMESTAMP({sql}))", (
                tzname,
                *params,
            )
        return sql, params

    def datetime_cast_date_sql(self, sql, params, tzname):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return f'({sql})::date', params

    def datetime_cast_time_sql(self, sql, params, tzname):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return f'({sql})::time', params

    def date_extract_sql(self, lookup_type, sql, params):
        # https://docs.snowflake.com/en/sql-reference/functions-date-time.html#label-supported-date-time-parts
        if lookup_type == 'week_day':
            # For consistency across backends, return Sunday=1, Saturday=7.
            return f"EXTRACT('dow', {sql}) + 1", params
        elif lookup_type == 'iso_week_day':
            return f"EXTRACT('dow_iso', {sql})", params
        elif lookup_type == 'iso_year':
            return f"EXTRACT('yearofweekiso', {sql})", params
        else:
            return f"EXTRACT(%s, {sql})", (lookup_type, *params)

    def datetime_extract_sql(self, lookup_type, sql, params, tzname):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return self.date_extract_sql(lookup_type, sql, params)

    def date_trunc_sql(self, lookup_type, sql, params, tzname=None):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return f"DATE_TRUNC(%s, {sql})", (lookup_type, *params)

    def datetime_trunc_sql(self, lookup_type, sql, params, tzname):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return f"DATE_TRUNC(%s, {sql})", (lookup_type, *params)

    def time_trunc_sql(self, lookup_type, sql, params, tzname=None):
        sql, params = self._convert_sql_to_tz(sql, params, tzname)
        return f"DATE_TRUNC(%s, {sql})::time", (lookup_type, *params)

    def format_for_duration_arithmetic(self, sql):
        return "INTERVAL '%s MICROSECONDS'" % sql

    def get_db_converters(self, expression):
        converters = super().get_db_converters(expression)
        internal_type = expression.output_field.get_internal_type()
        if internal_type == 'DateTimeField':
            if not settings.USE_TZ:
                converters.append(self.convert_datetimefield_value)
        elif internal_type == 'UUIDField':
            converters.append(self.convert_uuidfield_value)
        return converters

    def convert_datetimefield_value(self, value, expression, connection):
        if value is not None:
            # Django expects naive datetimes when settings.USE_TZ is False.
            value = timezone.make_naive(value)
        return value

    def convert_durationfield_value(self, value, expression, connection):
        # Snowflake sometimes returns Decimal which is an unsupported type for
        # timedelta microseconds component.
        if isinstance(value, decimal.Decimal):
            value = float(value)
        return super().convert_durationfield_value(value, expression, connection)

    def convert_uuidfield_value(self, value, expression, connection):
        if value is not None:
            value = uuid.UUID(value)
        return value

    def explain_query_prefix(self, format=None, **options):
        if format is None:
            format = 'TABULAR'
        prefix = super().explain_query_prefix(format, **options)
        return prefix + ' ' + format

    def last_executed_query(self, cursor, sql, params):
        return cursor.query

    def last_insert_id(self, cursor, table_name, pk_name):
        # This is subject to race conditions.
        return cursor.execute(
            'SELECT MAX({pk_name}) FROM {table_name}'.format(
                pk_name=self.quote_name(pk_name),
                table_name=self.quote_name(table_name),
            )
        ).fetchone()[0]

    def limit_offset_sql(self, low_mark, high_mark):
        # This method is copied from BaseDatabaseOperations with 'LIMIT %d'
        # replaced with 'LIMIT %s' to allow "LIMIT null" for no limit.
        limit, offset = self._get_limit_offset_params(low_mark, high_mark)
        return ' '.join(sql for sql in (
            ('LIMIT %s' % limit) if limit else None,
            ('OFFSET %d' % offset) if offset else None,
        ) if sql)

    def no_limit_value(self):
        return 'null'

    def quote_name(self, name):
        if name.startswith('"') and name.endswith('"'):
            return name  # Quoting once is enough.
        return '"%s"' % name.upper().replace('.', '"."')

    def regex_lookup(self, lookup_type):
        match_option = 'c' if lookup_type == 'regex' else 'i'
        return f"REGEXP_INSTR(%s, %s, 1, 1, 0, '{match_option}') > 0"

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        if not tables:
            return []
        sql = []
        if reset_sequences:
            sql.extend(
                '%s %s;' % (
                    style.SQL_KEYWORD('TRUNCATE'),
                    style.SQL_FIELD(self.quote_name(table_name)),
                ) for table_name in tables
            )
        else:
            # DELETE to preserve sequences.
            sql.extend(
                '%s %s %s;' % (
                    style.SQL_KEYWORD('DELETE'),
                    style.SQL_KEYWORD('FROM'),
                    style.SQL_FIELD(self.quote_name(table_name)),
                ) for table_name in tables
            )
        return sql

    def subtract_temporals(self, internal_type, lhs, rhs):
        lhs_sql, lhs_params = lhs
        rhs_sql, rhs_params = rhs
        if internal_type == 'TimeField':
            # Cast rhs_sql with TO_TIME in case it's a string.
            return f"TIMEDIFF(MICROSECOND, TO_TIME({rhs_sql}), {lhs_sql})", (*rhs_params, *lhs_params)
        return f"TIMEDIFF(MICROSECOND, {rhs_sql}, {lhs_sql})", (*rhs_params, *lhs_params)
